/********************************************************
 *  ██████╗  ██████╗████████╗██╗
 * ██╔════╝ ██╔════╝╚══██╔══╝██║
 * ██║  ███╗██║        ██║   ██║
 * ██║   ██║██║        ██║   ██║
 * ╚██████╔╝╚██████╗   ██║   ███████╗
 *  ╚═════╝  ╚═════╝   ╚═╝   ╚══════╝
 * Geophysical Computational Tools & Library (GCTL)
 *
 * Copyright (c) 2022  Yi Zhang (yizhang-geo@zju.edu.cn)
 *
 * GCTL is distributed under a dual licensing scheme. You can redistribute 
 * it and/or modify it under the terms of the GNU Lesser General Public 
 * License as published by the Free Software Foundation, either version 2 
 * of the License, or (at your option) any later version. You should have 
 * received a copy of the GNU Lesser General Public License along with this 
 * program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * If the terms and conditions of the LGPL v.2. would prevent you from using 
 * the GCTL, please consider the option to obtain a commercial license for a 
 * fee. These licenses are offered by the GCTL's original author. As a rule, 
 * licenses are provided "as-is", unlimited in time for a one time fee. Please 
 * send corresponding requests to: yizhang-geo@zju.edu.cn. Please do not forget 
 * to include some description of your company and the realm of its activities. 
 * Also add information on how to contact you by electronic and paper mail.
 ******************************************************/

#ifndef _GCTL_GRAV_KERNEL_TRIANGLE_H
#define _GCTL_GRAV_KERNEL_TRIANGLE_H

#include "gm_data.h"

namespace gctl
{
	/**
	 * @brief      三角形面元正演重力场需要的参数
	 */
	struct gravtri_para
	{
		tensor F; ///< 面的张量积
		tensor E[3]; ///< 边的张量积
		double edglen[3]; ///< 边的长度
	};

	typedef type_triangle<gravtri_para> grav_triangle; ///< 带gravtri_para属性的三角形结构体

	/**
	 * @brief      calculate the gravity parameters of given element type.
	 *
	 * @param      in_tri    Input elements
	 * @param      out_para  The output parameter array
	 */
	void callink_gravity_para(array<grav_triangle> &in_tri, array<gravtri_para> &out_para);

    /**
	 * @brief      正演计算直角坐标下的三维三角形单元体重力位
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维三角形单元体数组
	 * @param[in]  obsp           三维直角坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<double> &out_obs, const array<grav_triangle> &ele, const array<point3dc> &obsp, 
		double rho, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      正演计算直角坐标下的三维三角形单元体重力值
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维三角形单元体数组
	 * @param[in]  obsp           三维直角坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<point3dc> &out_obs, const array<grav_triangle> &ele, const array<point3dc> &obsp, 
		double rho, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      正演计算直角坐标下的三维三角形单元体重力梯度值
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维三角形单元体数组
	 * @param[in]  obsp           三维直角坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<tensor> &out_obs, const array<grav_triangle> &ele, const array<point3dc> &obsp, 
		double rho, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      正演计算球角坐标下的三维三角形单元体重力位
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维三角形单元体数组
	 * @param[in]  obsp           三维直角坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<double> &out_obs, const array<grav_triangle> &ele, const array<point3ds> &obsp, 
		double rho, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      正演计算球角坐标下的三维三角形单元体重力值
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维三角形单元体数组
	 * @param[in]  obsp           三维直角坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<point3dc> &out_obs, const array<grav_triangle> &ele, const array<point3ds> &obsp, 
		double rho, verbose_type_e verbose = FullMsg);

	/**
	 * @brief      正演计算球角坐标下的三维三角形单元体重力梯度值
	 *
	 * @param      out_obs        输出的重力观测值
	 * @param[in]  ele            三维三角形单元体数组
	 * @param[in]  obsp           三维直角坐标下的观测点数组
	 * @param[in]  rho            矩形单元体密度数组
	 * @param[in]  verbose        返回信息层级
	 */
	void gobser(array<tensor> &out_obs, const array<grav_triangle> &ele, const array<point3ds> &obsp, 
		double rho, verbose_type_e verbose = FullMsg);
}

#endif // _GCTL_GRAV_KERNEL_TRIANGLE_H